@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2019 - 2023, Ilan Schnell; All Rights Reserved
+   Copyright (c) 2019 - 2024, Ilan Schnell; All Rights Reserved
    bitarray is published under the PSF license.
 
    This file contains the C implementation of some useful utility functions.
@@ -42,16 +42,10 @@ new_bitarray(Py_ssize_t nbits, PyObject *endian, int c)
     PyObject *args;             /* args for bitarray() */
     bitarrayobject *res;
 
-    if ((args = PyTuple_New(3)) == NULL)  /* arguments for bitarray() */
+    args = Py_BuildValue("nOO", nbits, endian, Py_Ellipsis);
+    if (args == NULL)
         return NULL;
 
-    /* PyTuple_SET_ITEM "steals" a reference to item */
-    PyTuple_SET_ITEM(args, 0, PyLong_FromSsize_t(nbits));
-    Py_INCREF(endian);
-    PyTuple_SET_ITEM(args, 1, endian);
-    Py_INCREF(Py_Ellipsis);
-    PyTuple_SET_ITEM(args, 2, Py_Ellipsis);
-
     /* equivalent to: res = bitarray(nbits, endian, Ellipsis) */
     res = (bitarrayobject *) PyObject_CallObject(bitarray_type_obj, args);
     Py_DECREF(args);
@@ -492,14 +486,37 @@ hex_to_int(char c)
     return -1;
 }
 
+/* create hexadecimal string from bitarray */
+static char *
+ba2hex_core(bitarrayobject *a)
+{
+    const int le = IS_LE(a), be = IS_BE(a);
+    const size_t strsize = a->nbits / 4;
+    char *buff = a->ob_item, *str;
+    size_t i;
+
+    assert(a->nbits % 4 == 0);
+
+    str = (char *) PyMem_Malloc(strsize + 1);
+    if (str == NULL)
+        return NULL;
+
+    /* translate entire bitarray buffer, even when we have 4 padbits */
+    for (i = 0; i < strsize; i += 2) {
+        unsigned char c = *buff++;
+        str[i + le] = hexdigits[c >> 4];
+        str[i + be] = hexdigits[0x0f & c];
+    }
+    str[strsize] = 0;  /* terminate string */
+    return str;
+}
+
 static PyObject *
 ba2hex(PyObject *module, PyObject *obj)
 {
     PyObject *result;
     bitarrayobject *a;
-    size_t i, strsize;
     char *str;
-    int le, be;
 
     if (ensure_bitarray(obj) < 0)
         return NULL;
@@ -509,23 +526,14 @@ ba2hex(PyObject *module, PyObject *obj)
         return PyErr_Format(PyExc_ValueError, "bitarray length %zd not "
                             "multiple of 4", a->nbits);
 
-    /* We want strsize to be even, such that we can transform the entire
-       bitarray buffer at once.  Hence, we don't use a->nbits / 4 here, as
-       is could make strsize odd. */
-    strsize = 2 * Py_SIZE(obj);
-    str = (char *) PyMem_Malloc(strsize);
-    if (str == NULL)
+    if ((str = ba2hex_core(a)) == NULL)
         return PyErr_NoMemory();
 
-    le = IS_LE(a);
-    be = IS_BE(a);
-    for (i = 0; i < strsize; i += 2) {
-        unsigned char c = a->ob_item[i / 2];
-        str[i + le] = hexdigits[c >> 4];
-        str[i + be] = hexdigits[0x0f & c];
-    }
-    assert((size_t) a->nbits / 4 <= strsize);
-    result = Py_BuildValue("s#", str, a->nbits / 4);
+#if IS_PY3K
+    result = PyUnicode_FromString(str);
+#else
+    result = Py_BuildValue("s", str);
+#endif
     PyMem_Free((void *) str);
     return result;
 }
@@ -537,95 +545,57 @@ Return a string containing the hexadecimal representation of\n\
 the bitarray (which has to be multiple of 4 in length).");
 
 
-/* Translate hexadecimal digits 'bytes' into the bitarray 'a' buffer.
+/* Translate hexadecimal digits from 'hexstr' into the bitarray 'a' buffer.
    Each digit corresponds to 4 bits in the bitarray.
    Note that the number of hexadecimal digits may be odd. */
 static int
-hex2ba_core(bitarrayobject *a, PyObject *bytes)
+hex2ba_core(bitarrayobject *a, Py_buffer hexstr)
 {
-    Py_ssize_t strsize, i;
-    char *str;
-    int le = IS_LE(a), be = IS_BE(a);
+    const char *str = hexstr.buf;
+    const int be = IS_BE(a);
+    Py_ssize_t i;
 
-    assert(PyBytes_Check(bytes));
-    str = PyBytes_AS_STRING(bytes);
-    strsize = PyBytes_GET_SIZE(bytes);
-    assert(a->nbits == 4 * strsize);
-    assert(le + be == 1 && str[strsize] == 0);
+    assert(a->nbits == 4 * hexstr.len);
 
-    for (i = 0; i < strsize; i += 2) {
-        int x = hex_to_int(str[i + le]);
-        int y = hex_to_int(str[i + be]);
+    memset(a->ob_item, 0, Py_SIZE(a));
+    for (i = 0; i < hexstr.len; i++) {
+        unsigned char c = str[i];
+        int x = hex_to_int(c);
 
-        if (x < 0 || y < 0) {
-            /* ignore terminating NUL - happends when strsize is odd */
-            if (i + le == strsize) /* str[i+le] is NUL */
-                x = 0;
-            if (i + be == strsize) /* str[i+be] is NUL */
-                y = 0;
-            /* there is an invalid byte - or (non-terminating) NUL */
-            if (x < 0 || y < 0) {
-                PyErr_SetString(PyExc_ValueError,
-                                "non-hexadecimal digit found");
-                return -1;
-            }
+        if (x < 0) {
+            PyErr_Format(PyExc_ValueError, "non-hexadecimal digit found, "
+                         "got '%c' (0x%02x)", c, c);
+            return -1;
         }
-        assert(0 <= x && x < 16 && 0 <= y && y < 16);
-        a->ob_item[i / 2] = x << 4 | y;
+        assert(0 <= x && x < 16);
+        a->ob_item[i / 2] |= x << 4 * ((i + be) % 2);
     }
     return 0;
 }
 
-/* Return new reference to bytes object from either ASCII str (unicode) or
-   bytes.  On failure, set exception and return NULL. */
-static PyObject *
-anystr_to_bytes(PyObject *obj)
-{
-    PyObject *bytes;
-
-    if (PyUnicode_Check(obj)) {
-        bytes = PyUnicode_AsASCIIString(obj);
-        if (bytes == NULL)
-            return NULL;
-    }
-    else if (PyBytes_Check(obj)) {
-        bytes = obj;
-        Py_INCREF(bytes);
-    }
-    else {
-        PyErr_Format(PyExc_TypeError, "str or bytes expected, got '%s'",
-                     Py_TYPE(obj)->tp_name);
-        return NULL;
-    }
-    assert(PyBytes_Check(bytes));
-    return bytes;
-}
-
 static PyObject *
 hex2ba(PyObject *module, PyObject *args, PyObject *kwds)
 {
     static char *kwlist[] = {"", "endian", NULL};
-    PyObject *obj, *bytes, *endian = Py_None;
+    PyObject *endian = Py_None;
+    Py_buffer hexstr;
     bitarrayobject *a;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|O:hex2ba", kwlist,
-                                     &obj, &endian))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s*|O:hex2ba", kwlist,
+                                     &hexstr, &endian))
         return NULL;
 
-    if ((bytes = anystr_to_bytes(obj)) == NULL)
-        return NULL;
-
-    a = new_bitarray(4 * PyBytes_GET_SIZE(bytes), endian, -1);
+    a = new_bitarray(4 * hexstr.len, endian, -1);
     if (a == NULL)
         goto error;
 
-    if (hex2ba_core(a, bytes) < 0)  /* do translation here */
+    if (hex2ba_core(a, hexstr) < 0)
         goto error;
 
-    Py_DECREF(bytes);
+    PyBuffer_Release(&hexstr);
     return (PyObject *) a;
  error:
-    Py_DECREF(bytes);
+    PyBuffer_Release(&hexstr);
     Py_XDECREF((PyObject *) a);
     return NULL;
 }
@@ -641,7 +611,7 @@ Bitarray of hexadecimal representation.  hexstr may contain any number\n\
 /* RFC 4648 Base32 alphabet */
 static const char base32_alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
 
-/* standard base 64 alphabet */
+/* standard base 64 alphabet - also described on RFC 4648 */
 static const char base64_alphabet[] =
     "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
@@ -694,47 +664,69 @@ base_to_length(int n)
     return -1;
 }
 
+/* create ASCII string from bitarray and base length m */
+static char *
+ba2base_core(bitarrayobject *a, int m)
+{
+    const int le = IS_LE(a);
+    const size_t strsize = a->nbits / m;
+    const char *alphabet;
+    size_t i = 0, j;
+    char *str;
+
+    assert(1 <= m && m <= 6 && a->nbits % m == 0);
+
+    switch (m) {
+    case 5: alphabet = base32_alphabet; break;
+    case 6: alphabet = base64_alphabet; break;
+    default: alphabet = hexdigits;
+    }
+
+    str = (char *) PyMem_Malloc(strsize + 1);
+    if (str == NULL)
+        return NULL;
+
+    for (j = 0; j < strsize; j++) {
+        int k, x = 0;
+
+        for (k = 0; k < m; k++) {
+            int q = le ? k : (m - k - 1);
+            x |= getbit(a, i++) << q;
+        }
+        str[j] = alphabet[x];
+    }
+    str[strsize] = 0;  /* terminate string */
+    return str;
+}
+
 static PyObject *
 ba2base(PyObject *module, PyObject *args)
 {
-    const char *alphabet;
     bitarrayobject *a;
     PyObject *result;
-    size_t i, strsize;
     char *str;
-    int n, m, le;
+    int n, m;
 
     if (!PyArg_ParseTuple(args, "iO!:ba2base", &n,
                           bitarray_type_obj, (PyObject *) &a))
         return NULL;
+
     if ((m = base_to_length(n)) < 0)
         return NULL;
 
-    switch (n) {
-    case 32: alphabet = base32_alphabet; break;
-    case 64: alphabet = base64_alphabet; break;
-    default: alphabet = hexdigits;
-    }
-
     if (a->nbits % m)
         return PyErr_Format(PyExc_ValueError,
                             "bitarray length must be multiple of %d", m);
 
-    strsize = a->nbits / m;
-    if ((str = (char *) PyMem_Malloc(strsize)) == NULL)
+    str = m == 4 ? ba2hex_core(a) : ba2base_core(a, m);
+    if (str == NULL)
         return PyErr_NoMemory();
 
-    le = IS_LE(a);
-    for (i = 0; i < strsize; i++) {
-        int j, x = 0;
-
-        for (j = 0; j < m; j++) {
-            int k = le ? j : (m - j - 1);
-            x |= getbit(a, i * m + k) << j;
-        }
-        str[i] = alphabet[x];
-    }
-    result = Py_BuildValue("s#", str, strsize);
+#if IS_PY3K
+    result = PyUnicode_FromString(str);
+#else
+    result = Py_BuildValue("s", str);
+#endif
     PyMem_Free((void *) str);
     return result;
 }
@@ -745,42 +737,32 @@ PyDoc_STRVAR(ba2base_doc,
 Return a string containing the base `n` ASCII representation of\n\
 the bitarray.  Allowed values for `n` are 2, 4, 8, 16, 32 and 64.\n\
 The bitarray has to be multiple of length 1, 2, 3, 4, 5 or 6 respectively.\n\
-For `n=16` (hexadecimal), `ba2hex()` will be much faster, as `ba2base()`\n\
-does not take advantage of byte level operations.\n\
 For `n=32` the RFC 4648 Base32 alphabet is used, and for `n=64` the\n\
 standard base 64 alphabet is used.");
 
 
-/* Translate ASCII digits 'bytes' into bitarray buffer.
-   Arguments to this functions are:
-   - bitarray (of length m * len(s)) whose elements are overwritten
-   - byte object s containing the ASCII digits
-   - bits per digit - the base length m  [1..6]
-*/
+/* translate ASCII digits (with base length m) into bitarray buffer */
 static int
-base2ba_core(bitarrayobject *a, PyObject *bytes, int m)
+base2ba_core(bitarrayobject *a, Py_buffer asciistr, int m)
 {
-    Py_ssize_t strsize, i;
-    char *str;
-    int le = IS_LE(a), n = 1 << m;
+    const char *str = asciistr.buf;
+    const int le = IS_LE(a), n = 1 << m;
+    Py_ssize_t i = 0, j;
 
-    assert(PyBytes_Check(bytes));
-    str = PyBytes_AS_STRING(bytes);
-    strsize = PyBytes_GET_SIZE(bytes);
-    assert(a->nbits == m * strsize);
+    assert(a->nbits == asciistr.len * m && 1 <= m && m <= 6);
 
-    for (i = 0; i < strsize; i++) {
-        int j, d = digit_to_int(n, str[i]);
+    for (j = 0; j < asciistr.len; j++) {
+        unsigned char c = str[j];
+        int k, x = digit_to_int(n, c);
 
-        if (d < 0) {
-            unsigned char c = str[i];
+        if (x < 0) {
             PyErr_Format(PyExc_ValueError, "invalid digit found for "
                          "base %d, got '%c' (0x%02x)", n, c, c);
             return -1;
         }
-        for (j = 0; j < m; j++) {
-            int k = le ? j : (m - j - 1);
-            setbit(a, i * m + k, d & (1 << j));
+        for (k = 0; k < m; k++) {
+            int q = le ? k : (m - k - 1);
+            setbit(a, i++, x & (1 << q));
         }
     }
     return 0;
@@ -790,31 +772,30 @@ static PyObject *
 base2ba(PyObject *module, PyObject *args, PyObject *kwds)
 {
     static char *kwlist[] = {"", "", "endian", NULL};
-    PyObject *obj, *bytes, *endian = Py_None;
-    bitarrayobject *a;
-    int n, m;                   /* n = 2^m */
+    PyObject *endian = Py_None;
+    Py_buffer asciistr;
+    bitarrayobject *a = NULL;
+    int m, n, t;                   /* n = 2^m */
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "iO|O:base2ba", kwlist,
-                                     &n, &obj, &endian))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "is*|O:base2ba", kwlist,
+                                     &n, &asciistr, &endian))
         return NULL;
 
     if ((m = base_to_length(n)) < 0)
-        return NULL;
+        goto error;
 
-    if ((bytes = anystr_to_bytes(obj)) == NULL)
-        return NULL;
-
-    a = new_bitarray(m * PyBytes_GET_SIZE(bytes), endian, -1);
+    a = new_bitarray(m * asciistr.len, endian, -1);
     if (a == NULL)
         goto error;
 
-    if (base2ba_core(a, bytes, m) < 0)  /* do translation here */
+    t = m == 4 ? hex2ba_core(a, asciistr) : base2ba_core(a, asciistr, m);
+    if (t < 0)
         goto error;
 
-    Py_DECREF(bytes);
+    PyBuffer_Release(&asciistr);
     return (PyObject *) a;
  error:
-    Py_DECREF(bytes);
+    PyBuffer_Release(&asciistr);
     Py_XDECREF((PyObject *) a);
     return NULL;
 }
@@ -824,8 +805,6 @@ PyDoc_STRVAR(base2ba_doc,
 \n\
 Bitarray of base `n` ASCII representation.\n\
 Allowed values for `n` are 2, 4, 8, 16, 32 and 64.\n\
-For `n=16` (hexadecimal), `hex2ba()` will be much faster, as `base2ba()`\n\
-does not take advantage of byte level operations.\n\
 For `n=32` the RFC 4648 Base32 alphabet is used, and for `n=64` the\n\
 standard base 64 alphabet is used.");
 
@@ -916,8 +895,8 @@ next_char(PyObject *iter)
     return (int) c;
 }
 
-/* write n bytes (into buffer str) representing integer i (using
-   little endian byte-order) */
+/* write n bytes (into buffer str) representing non-negative integer i,
+   using little endian byte-order */
 static void
 write_n(char *str, int n, Py_ssize_t i)
 {
@@ -931,7 +910,7 @@ write_n(char *str, int n, Py_ssize_t i)
     assert(i == 0);
 }
 
-/* read n bytes from iter and return corresponding positive integer,
+/* read n bytes from iter and return corresponding non-negative integer,
    using little endian byte-order */
 static Py_ssize_t
 read_n(int n, PyObject *iter)
@@ -953,7 +932,7 @@ read_n(int n, PyObject *iter)
     return i;
 }
 
-/* return number of bytes necessary to represent i */
+/* return number of bytes necessary to represent non-negative integer i */
 static int
 byte_length(Py_ssize_t i)
 {
